from __future__ import annotations

from typing import TYPE_CHECKING

from InquirerPy import inquirer
from InquirerPy.base.control import Choice
from InquirerPy.validator import EmptyInputValidator
from rich.console import Console

if TYPE_CHECKING:
    from cyberdrop_dl.managers.manager import Manager

console = Console()


def edit_global_settings_prompt(manager: Manager) -> None:
    """Edit the authentication values"""
    while True:
        console.clear()
        console.print("Editing Global Settings")
        action = inquirer.select(
            message="What would you like to do?",
            choices=[
                Choice(1, "Edit General Settings"),
                Choice(2, "Edit Rate Limiting Settings"),
                Choice(3, "Edit UI Options"),
                Choice(4, "Done"),
            ],
            vi_mode=manager.vi_mode,
        ).execute()

        # Edit General Settings
        if action == 1:
            edit_general_settings_prompt(manager)

        # Edit Rate Limiting Settings
        elif action == 2:
            edit_rate_limiting_settings_prompt(manager)
            
        # Edit UI Settings
        elif action == 3:
            edit_ui_settings_prompt(manager)

        # Done
        elif action == 4:
            manager.config_manager.write_updated_global_settings_config()
            break


def edit_general_settings_prompt(manager: Manager) -> None:
    """Edit the general settings"""
    console.clear()
    console.print("Editing General Settings")
    allow_insecure_connections = inquirer.confirm("Allow insecure connections?", vi_mode=manager.vi_mode).execute()
    user_agent = inquirer.text(
        message="User Agent:",
        default=manager.config_manager.global_settings_data['General']['user_agent'],
        validate=EmptyInputValidator("Input should not be empty"),
        vi_mode=manager.vi_mode,
    ).execute()
    proxy = inquirer.text(
        message="Proxy:",
        default=manager.config_manager.global_settings_data['General']['proxy'],
        vi_mode=manager.vi_mode,
    ).execute()
    flaresolverr = inquirer.text(
        message="FlareSolverr (IP:PORT):",
        default=manager.config_manager.global_settings_data['General']['flaresolverr'],
        vi_mode=manager.vi_mode,
    ).execute()
    max_filename_length = inquirer.number(
        message="Max Filename Length:",
        default=int(manager.config_manager.global_settings_data['General']['max_file_name_length']),
        float_allowed=False,
        vi_mode=manager.vi_mode,
    ).execute()
    max_folder_name_length = inquirer.number(
        message="Max Folder Name Length:",
        default=int(manager.config_manager.global_settings_data['General']['max_folder_name_length']),
        float_allowed=False,
        vi_mode=manager.vi_mode,
    ).execute()
    required_free_space = inquirer.number(
        message="Required Free Space (in GB):",
        default=int(manager.config_manager.global_settings_data['General']['required_free_space']),
        float_allowed=False,
        vi_mode=manager.vi_mode,
    ).execute()

    manager.config_manager.global_settings_data['General']['allow_insecure_connections'] = allow_insecure_connections
    manager.config_manager.global_settings_data['General']['user_agent'] = user_agent
    manager.config_manager.global_settings_data['General']['proxy'] = proxy
    manager.config_manager.global_settings_data['General']['flaresolverr'] = flaresolverr
    manager.config_manager.global_settings_data['General']['max_filename_length'] = int(max_filename_length)
    manager.config_manager.global_settings_data['General']['max_folder_name_length'] = int(max_folder_name_length)
    manager.config_manager.global_settings_data['General']['required_free_space'] = int(required_free_space)


def edit_ui_settings_prompt(manager: Manager) -> None:
    """Edit the progress settings"""
    console.clear()
    console.print("Editing UI Settings")
    refresh_rate = inquirer.number(
        message="Refresh Rate:",
        default=int(manager.config_manager.global_settings_data['UI_Options']['refresh_rate']),
        float_allowed=False,
        vi_mode=manager.vi_mode,
    ).execute()
    scraping_item_limit = inquirer.number(
        message="Number of lines to allow for scraping items before overflow:",
        default=int(manager.config_manager.global_settings_data['UI_Options']['scraping_item_limit']),
        float_allowed=False,
        vi_mode=manager.vi_mode,
    ).execute()
    downloading_item_limit = inquirer.number(
        message="Number of lines to allow for download items before overflow:",
        default=int(manager.config_manager.global_settings_data['UI_Options']['downloading_item_limit']),
        float_allowed=False,
        vi_mode=manager.vi_mode,
    ).execute()
    
    vi_mode = inquirer.confirm(
        message="Enable VI/VIM keybindings?",
        default=bool(manager.config_manager.global_settings_data['UI_Options']['vi_mode']),
        vi_mode=manager.vi_mode,
    ).execute()

    manager.config_manager.global_settings_data['UI_Options']['refresh_rate'] = int(refresh_rate)
    manager.config_manager.global_settings_data['UI_Options']['scraping_item_limit'] = int(scraping_item_limit)
    manager.config_manager.global_settings_data['UI_Options']['downloading_item_limit'] = int(downloading_item_limit)
    manager.config_manager.global_settings_data['UI_Options']['vi_mode'] = vi_mode
    manager.vi_mode = vi_mode


def edit_rate_limiting_settings_prompt(manager: Manager) -> None:
    """Edit the rate limiting settings"""
    console.clear()
    console.print("Editing Rate Limiting Settings")
    connection_timeout = inquirer.number(
        message="Connection Timeout (in seconds):",
        default=int(manager.config_manager.global_settings_data['Rate_Limiting_Options']['connection_timeout']),
        float_allowed=False,
        vi_mode=manager.vi_mode,
    ).execute()
    read_timeout = inquirer.number(
        message="Read Timeout (in seconds):",
        default=int(manager.config_manager.global_settings_data['Rate_Limiting_Options']['read_timeout']),
        float_allowed=False,
        vi_mode=manager.vi_mode,
    ).execute()
    download_attempts = inquirer.number(
        message="Download Attempts:",
        default=int(manager.config_manager.global_settings_data['Rate_Limiting_Options']['download_attempts']),
        float_allowed=False,
        vi_mode=manager.vi_mode,
    ).execute()
    rate_limit = inquirer.number(
        message="Maximum number of requests per second:",
        default=int(manager.config_manager.global_settings_data['Rate_Limiting_Options']['rate_limit']),
        float_allowed=False,
        vi_mode=manager.vi_mode,
    ).execute()
    throttle = inquirer.number(
        message="Delay between requests during the download stage:",
        default=float(manager.config_manager.global_settings_data['Rate_Limiting_Options']['download_delay']),
        float_allowed=True,
        vi_mode=manager.vi_mode,
    ).execute()

    max_simultaneous_downloads = inquirer.number(
        message="Maximum number of simultaneous downloads:",
        default=int(manager.config_manager.global_settings_data['Rate_Limiting_Options']['max_simultaneous_downloads']),
        float_allowed=False,
        vi_mode=manager.vi_mode,
    ).execute()
    max_simultaneous_downloads_per_domain = inquirer.number(
        message="Maximum number of simultaneous downloads per domain:",
        default=int(manager.config_manager.global_settings_data['Rate_Limiting_Options']['max_simultaneous_downloads_per_domain']),
        float_allowed=False,
        vi_mode=manager.vi_mode,
    ).execute()

    manager.config_manager.global_settings_data['Rate_Limiting_Options']['connection_timeout'] = int(connection_timeout)
    manager.config_manager.global_settings_data['Rate_Limiting_Options']['read_timeout'] = int(read_timeout)
    manager.config_manager.global_settings_data['Rate_Limiting_Options']['download_attempts'] = int(download_attempts)
    manager.config_manager.global_settings_data['Rate_Limiting_Options']['rate_limit'] = int(rate_limit)
    manager.config_manager.global_settings_data['Rate_Limiting_Options']['download_delay'] = float(throttle)
    manager.config_manager.global_settings_data['Rate_Limiting_Options']['max_simultaneous_downloads'] = int(max_simultaneous_downloads)
    manager.config_manager.global_settings_data['Rate_Limiting_Options']['max_simultaneous_downloads_per_domain'] = int(max_simultaneous_downloads_per_domain)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from InquirerPy import inquirer
from InquirerPy.validator import PathValidator
from rich.console import Console

from cyberdrop_dl import __version__
from cyberdrop_dl.ui.prompts.settings_authentication_prompts import edit_authentication_values_prompt
from cyberdrop_dl.ui.prompts.general_prompts import (
    main_prompt, select_config_prompt, donations_prompt,
    import_cyberdrop_v4_items_prompt, manage_configs_prompt)
from cyberdrop_dl.ui.prompts.settings_global_prompts import edit_global_settings_prompt
from cyberdrop_dl.ui.prompts.url_file_prompts import edit_urls_prompt
from cyberdrop_dl.ui.prompts.settings_user_prompts import create_new_config_prompt, edit_config_values_prompt

console = Console()

if TYPE_CHECKING:
    from cyberdrop_dl.managers.manager import Manager


def program_ui(manager: Manager):
    """Program UI"""
    while True:
        console.clear()
        console.print(f"[bold]Cyberdrop Downloader (V{str(__version__)})[/bold]")
        console.print(f"[bold]Current Config:[/bold] {manager.config_manager.loaded_config}")

        action = main_prompt(manager)

        # Download
        if action == 1:
            break

        # Download (All Configs)
        if action == 2:
            manager.args_manager.all_configs = True
            break

        # Retry Failed Downloads
        elif action == 3:
            manager.args_manager.retry = True
            break
            
        # Sort All Configs
        elif action == 4:
            manager.args_manager.sort_all_configs = True
            manager.args_manager.all_configs = True
            break

        # Edit URLs
        elif action == 5:
            input_file = manager.config_manager.settings_data['Files']['input_file'] if not manager.args_manager.input_file else manager.args_manager.input_file
            edit_urls_prompt(input_file, manager.vi_mode)

        # Select Config
        elif action == 6:
            configs = manager.config_manager.get_configs()
            selected_config = select_config_prompt(manager, configs)
            manager.config_manager.change_config(selected_config)

        elif action == 7:
            console.clear()
            console.print("Editing Input / Output File Paths")
            input_file = inquirer.filepath(
                message="Enter the input file path:",
                default=str(manager.config_manager.settings_data['Files']['input_file']),
                validate=PathValidator(is_file=True, message="Input is not a file"),
                vi_mode=manager.vi_mode,
            ).execute()
            download_folder = inquirer.text(
                message="Enter the download folder path:",
                default=str(manager.config_manager.settings_data['Files']['download_folder']),
                validate=PathValidator(is_dir=True, message="Input is not a directory"),
                vi_mode=manager.vi_mode,
            ).execute()

            manager.config_manager.settings_data['Files']['input_file'] = Path(input_file)
            manager.config_manager.settings_data['Files']['download_folder'] = Path(download_folder)
            manager.config_manager.write_updated_settings_config()

        # Manage Configs
        elif action == 8:
            while True:
                console.clear()
                console.print("[bold]Manage Configs[/bold]")
                console.print(f"[bold]Current Config:[/bold] {manager.config_manager.loaded_config}")

                action = manage_configs_prompt(manager)

                # Change Default Config
                if action == 1:
                    configs = manager.config_manager.get_configs()
                    selected_config = select_config_prompt(manager, configs)
                    manager.config_manager.change_default_config(selected_config)

                # Create A Config
                elif action == 2:
                    create_new_config_prompt(manager)

                # Delete A Config
                elif action == 3:
                    configs = manager.config_manager.get_configs()
                    if len(configs) != 1:
                        selected_config = select_config_prompt(manager, configs)
                        if selected_config == manager.config_manager.loaded_config:
                            inquirer.confirm(
                                message="You cannot delete the currently active config, press enter to continue.",
                                default=False,
                                vi_mode=manager.vi_mode,
                            ).execute()
                            continue
                        manager.config_manager.delete_config(selected_config)
                    else:
                        inquirer.confirm(
                            message="There is only one config, press enter to continue.",
                            default=False,
                            vi_mode=manager.vi_mode,
                        ).execute()

                # Edit Config
                elif action == 4:
                    edit_config_values_prompt(manager)

                # Edit Authentication Values
                elif action == 5:
                    edit_authentication_values_prompt(manager)

                # Edit Global Settings
                elif action == 6:
                    edit_global_settings_prompt(manager)

                # Done
                elif action == 7:
                    break

        # Import Cyberdrop_V4 Items
        elif action == 9:
            import_cyberdrop_v4_items_prompt(manager)

        # Exit
        elif action == 10:
            exit(0)



from __future__ import annotations

import calendar
import datetime
from typing import TYPE_CHECKING

from aiolimiter import AsyncLimiter
from yarl import URL

from cyberdrop_dl.scraper.crawler import Crawler
from cyberdrop_dl.utils.dataclasses.url_objects import ScrapeItem
from cyberdrop_dl.utils.utilities import get_filename_and_ext, error_handling_wrapper

if TYPE_CHECKING:
    from cyberdrop_dl.managers.manager import Manager


class Rule34VaultCrawler(Crawler):
    def __init__(self, manager: Manager):
        super().__init__(manager, "rule34vault", "Rule34Vault")
        self.primary_base_url = URL("https://rule34vault.com")
        self.request_limiter = AsyncLimiter(10, 1)

    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""

    async def fetch(self, scrape_item: ScrapeItem) -> None:
        """Determines where to send the scrape item based on the url"""
        task_id = await self.scraping_progress.add_task(scrape_item.url)

        if "post" in scrape_item.url.parts:
            await self.file(scrape_item)
        elif "playlists" in scrape_item.url.parts:
            await self.playlist(scrape_item)
        else:
            await self.tag(scrape_item)

        await self.scraping_progress.remove_task(task_id)

    @error_handling_wrapper
    async def tag(self, scrape_item: ScrapeItem) -> None:
        """Scrapes an album"""
        async with self.request_limiter:
            soup = await self.client.get_BS4(self.domain, scrape_item.url)

        title = await self.create_title(scrape_item.url.parts[1], None, None)

        content_block = soup.select_one('div[class="grid ng-star-inserted"]')
        content = content_block.select('a[class="box ng-star-inserted"]')
        for file_page in content:
            link = file_page.get('href')
            if link.startswith("/"):
                link = f"{self.primary_base_url}{link}"
            link = URL(link)
            new_scrape_item = await self.create_scrape_item(scrape_item, link, title, True)
            self.manager.task_group.create_task(self.run(new_scrape_item))
        if not content:
            return

        if len(scrape_item.url.parts) > 2:
            page = int(scrape_item.url.parts[-1])
            next_page = scrape_item.url.with_path(f"/{scrape_item.url.parts[1]}/page/{page + 1}")
        else:
            next_page = scrape_item.url.with_path(f"/{scrape_item.url.parts[1]}/page/2")
        new_scrape_item = await self.create_scrape_item(scrape_item, next_page, "")
        self.manager.task_group.create_task(self.run(new_scrape_item))

    @error_handling_wrapper
    async def playlist(self, scrape_item: ScrapeItem) -> None:
        """Scrapes a playlist"""
        async with self.request_limiter:
            soup = await self.client.get_BS4(self.domain, scrape_item.url)

        title_str = soup.select_one('div[class*=title]').text
        title = await self.create_title(title_str, scrape_item.url.parts[-1], None)

        content_block = soup.select_one('div[class="grid ng-star-inserted"]')
        content = content_block.select('a[class="box ng-star-inserted"]')
        for file_page in content:
            link = file_page.get('href')
            if link.startswith("/"):
                link = f"{self.primary_base_url}{link}"
            link = URL(link)
            new_scrape_item = await self.create_scrape_item(scrape_item, link, title, True)
            self.manager.task_group.create_task(self.run(new_scrape_item))
        if not content:
            return

        if scrape_item.url.query:
            page = scrape_item.url.query.get("page")
            next_page = scrape_item.url.with_query({"page": int(page) + 1})
        else:
            next_page = scrape_item.url.with_query({"page": 2})
        new_scrape_item = await self.create_scrape_item(scrape_item, next_page, "")
        self.manager.task_group.create_task(self.run(new_scrape_item))

    @error_handling_wrapper
    async def file(self, scrape_item: ScrapeItem) -> None:
        """Scrapes an image"""
        async with self.request_limiter:
            soup = await self.client.get_BS4(self.domain, scrape_item.url)

        date = await self.parse_datetime(soup.select_one('div[class="text-primary ng-star-inserted"]').text.split("(")[1].split(")")[0])
        scrape_item.date = date

        image = soup.select_one('img[class*="img ng-star-inserted"]')
        if image:
            link = image.get('src')
            if link.startswith("/"):
                link = f"{self.primary_base_url}{link}"
            link = URL(link)
            filename, ext = await get_filename_and_ext(link.name)
            await self.handle_file(link, scrape_item, filename, ext)
        video = soup.select_one("video source")
        if video:
            link = video.get('src')
            if link.startswith("/"):
                link = f"{self.primary_base_url}{link}"
            link = URL(link)
            filename, ext = await get_filename_and_ext(link.name)
            await self.handle_file(link, scrape_item, filename, ext)

    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""

    async def parse_datetime(self, date: str) -> int:
        """Parses a datetime string into a unix timestamp"""
        date = datetime.datetime.strptime(date, "%b %d, %Y, %I:%M:%S %p")
        return calendar.timegm(date.timetuple())



from __future__ import annotations

import asyncio
import os
import traceback
from dataclasses import field, Field
from functools import wraps
from http import HTTPStatus
from pathlib import Path
from typing import TYPE_CHECKING

import aiohttp
import filedate

from cyberdrop_dl.clients.download_client import is_4xx_client_error
from cyberdrop_dl.clients.errors import DownloadFailure, InvalidContentTypeFailure, DDOSGuardFailure
from cyberdrop_dl.utils.utilities import CustomHTTPStatus, log

if TYPE_CHECKING:
    from cyberdrop_dl.clients.download_client import DownloadClient
    from cyberdrop_dl.managers.manager import Manager
    from cyberdrop_dl.utils.dataclasses.url_objects import MediaItem


def retry(f):
    """This function is a wrapper that handles retrying for failed downloads"""
    @wraps(f)
    async def wrapper(self, *args, **kwargs):
        while True:
            try:
                return await f(self, *args, **kwargs)
            except DownloadFailure as e:
                media_item = args[0]
                await self.attempt_task_removal(media_item)

                if e.status != 999:
                    media_item.current_attempt += 1

                if not self.manager.config_manager.settings_data['Download_Options']['disable_download_attempt_limit']:
                    if media_item.current_attempt >= self.manager.config_manager.global_settings_data['Rate_Limiting_Options']['download_attempts']:
                        if hasattr(e, "status"):
                            await self.manager.progress_manager.download_stats_progress.add_failure(e.status)
                            if hasattr(e, "message"):
                                await log(f"Download Failed: {media_item.url} with status {e.status} and message {e.message}", 40)
                                await self.manager.log_manager.write_download_error_log(media_item.url, f" {e.status} - {e.message}")
                            else:
                                await log(f"Download Failed: {media_item.url} with status {e.status}", 40)
                                await self.manager.log_manager.write_download_error_log(media_item.url, f" {e.status}")
                        else:
                            await self.manager.progress_manager.download_stats_progress.add_failure("Unknown")
                            await self.manager.log_manager.write_download_error_log(media_item.url, " See Log for Details")
                            await log(f"Download Failed: {media_item.url} with error {e}", 40)
                        await self.manager.progress_manager.download_progress.add_failed()
                        break

                if hasattr(e, "status"):
                    if hasattr(e, "message"):
                        await log(f"Download Failed: {media_item.url} with status {e.status} and message {e.message}", 40)
                    else:
                        await log(f"Download Failed: {media_item.url} with status {e.status}", 40)
                else:
                    await log(f"Download Failed: {media_item.url} with error {e}", 40)
                await log(f"Download Retrying: {media_item.url} with attempt {media_item.current_attempt}", 20)
            
            except DDOSGuardFailure as e:
                media_item = args[0]
                await self.attempt_task_removal(media_item)
                await log(f"Download Failed: {media_item.url} with error {e}", 40)
                await self.manager.log_manager.write_download_error_log(media_item.url, " DDOSGuard")
                await log(traceback.format_exc(), 40)
                await self.manager.progress_manager.download_stats_progress.add_failure("DDOSGuard")
                await self.manager.progress_manager.download_progress.add_failed()
                break
            
            except InvalidContentTypeFailure as e:
                media_item = args[0]
                await self.attempt_task_removal(media_item)
                await log(f"Download Failed: {media_item.url} received Invalid Content", 40)
                await self.manager.log_manager.write_download_error_log(media_item.url, "Invalid Content Received")
                await log(e.message, 40)
                await self.manager.progress_manager.download_stats_progress.add_failure("Invalid Content Type")
                await self.manager.progress_manager.download_progress.add_failed()
                break
            
            except Exception as e:
                media_item = args[0]
                await log(f"Download Failed: {media_item.url} with error {e}", 40)
                await self.attempt_task_removal(media_item)
                await log(traceback.format_exc(), 40)
                await self.manager.log_manager.write_download_error_log(media_item.url, " See Log For Details")
                await self.manager.progress_manager.download_stats_progress.add_failure("Unknown")
                await self.manager.progress_manager.download_progress.add_failed()
                break
    return wrapper


class Downloader:
    def __init__(self, manager: Manager, domain: str):
        self.manager: Manager = manager
        self.domain: str = domain

        self.client: DownloadClient = field(init=False)

        self._file_lock = manager.download_manager.file_lock
        self._semaphore: asyncio.Semaphore = field(init=False)

        self._additional_headers = {}

        self.processed_items: list = []
        self.waiting_items = 0
        self._current_attempt_filesize = {}

    async def startup(self) -> None:
        """Starts the downloader"""
        self.client = self.manager.client_manager.downloader_session
        self._semaphore = asyncio.Semaphore(await self.manager.download_manager.get_download_limit(self.domain))

        self.manager.path_manager.download_dir.mkdir(parents=True, exist_ok=True)
        if self.manager.config_manager.settings_data['Sorting']['sort_downloads']:
            self.manager.path_manager.sorted_dir.mkdir(parents=True, exist_ok=True)

    async def run(self, media_item: MediaItem) -> None:
        """Runs the download loop"""
        self.waiting_items += 1
        media_item.current_attempt = 0

        await self._semaphore.acquire()
        self.waiting_items -= 1
        if media_item.url.path not in self.processed_items:
            self.processed_items.append(media_item.url.path)
            await self.manager.progress_manager.download_progress.update_total()

            await log(f"Download Starting: {media_item.url}", 20)
            async with self.manager.client_manager.download_session_limit:
                try:
                    if isinstance(media_item.file_lock_reference_name, Field):
                        media_item.file_lock_reference_name = media_item.filename
                    await self._file_lock.check_lock(media_item.file_lock_reference_name)
                    
                    await self.download(media_item)
                except Exception as e:
                    await log(f"Download Failed: {media_item.url} with error {e}", 40)
                    await log(traceback.format_exc(), 40)
                    await self.manager.progress_manager.download_stats_progress.add_failure("Unknown")
                    await self.manager.progress_manager.download_progress.add_failed()
                else:
                    await log(f"Download Finished: {media_item.url}", 20)
                finally:
                    await self._file_lock.release_lock(media_item.file_lock_reference_name)
        self._semaphore.release()

    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""

    async def check_file_can_download(self, media_item: MediaItem) -> bool:
        """Checks if the file can be downloaded"""
        if not await self.manager.download_manager.check_free_space():
            await log(f"Download Skip {media_item.url} due to insufficient free space", 10)
            return False
        if not await self.manager.download_manager.check_allowed_filetype(media_item):
            await log(f"Download Skip {media_item.url} due to filetype restrictions", 10)
            return False
        return True

    async def set_file_datetime(self, media_item: MediaItem, complete_file: Path) -> None:
        """Sets the file's datetime"""
        if self.manager.config_manager.settings_data['Download_Options']['disable_file_timestamps']:
            return
        if not isinstance(media_item.datetime, Field):
            file = filedate.File(str(complete_file))
            file.set(
                created=media_item.datetime,
                modified=media_item.datetime,
                accessed=media_item.datetime,
            )
            
    async def attempt_task_removal(self, media_item: MediaItem) -> None:
        """Attempts to remove the task from the progress bar"""
        if not isinstance(media_item.task_id, Field):
            try:
                await self.manager.progress_manager.file_progress.remove_file(media_item.task_id)
            except ValueError:
                pass
        media_item.task_id = field(init=False)
            
    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""

    @retry
    async def download(self, media_item: MediaItem) -> None:
        """Downloads the media item"""
        if not await self.check_file_can_download(media_item):
            await self.manager.progress_manager.download_progress.add_skipped()
            return

        try:
            if not isinstance(media_item.current_attempt, int):
                media_item.current_attempt = 1

            if not await self.check_file_can_download(media_item):
                await self.manager.progress_manager.download_progress.add_skipped()
                return

            downloaded = await self.client.download_file(self.manager, self.domain, media_item)
            
            if downloaded:
                os.chmod(media_item.complete_file, 0o666)
                await self.set_file_datetime(media_item, media_item.complete_file)
                await self.attempt_task_removal(media_item)
                await self.manager.progress_manager.download_progress.add_completed()

        except (aiohttp.ClientPayloadError, aiohttp.ClientOSError, aiohttp.ClientResponseError, ConnectionResetError,
                DownloadFailure, FileNotFoundError, PermissionError, aiohttp.ServerDisconnectedError, 
                asyncio.TimeoutError, aiohttp.ServerTimeoutError) as e:
            if hasattr(e, "status"):
                if ((await is_4xx_client_error(e.status) and e.status != HTTPStatus.TOO_MANY_REQUESTS)
                        or e.status == HTTPStatus.SERVICE_UNAVAILABLE
                        or e.status == HTTPStatus.BAD_GATEWAY
                        or e.status == CustomHTTPStatus.WEB_SERVER_IS_DOWN):
                    await self.manager.progress_manager.download_progress.add_failed()
                    await self.manager.progress_manager.download_stats_progress.add_failure(e.status)
                    await self.attempt_task_removal(media_item)
                    if hasattr(e, "message"):
                        if not e.message:
                            e.message = "Download Failed"
                        await log(f"Download failed: {media_item.url} with status {e.status} and message {e.message}", 40)
                        await self.manager.log_manager.write_download_error_log(media_item.url, f" {e.status} - {e.message}")
                    else:
                        await log(f"Download Failed: {media_item.url} with status {e.status}", 40)
                        await self.manager.log_manager.write_download_error_log(media_item.url, f" {e.status}")
                    return

            if isinstance(media_item.partial_file, Path) and media_item.partial_file.is_file():
                size = media_item.partial_file.stat().st_size
                if media_item.filename in self._current_attempt_filesize and self._current_attempt_filesize[media_item.filename] >= size:
                    raise DownloadFailure(status=getattr(e, "status", type(e).__name__), message="Download failed")
                self._current_attempt_filesize[media_item.filename] = size
                media_item.current_attempt = 0
                raise DownloadFailure(status=999, message="Download timeout reached, retrying")

            raise DownloadFailure(status=getattr(e, "status", type(e).__name__), message=repr(e))



from __future__ import annotations

import asyncio
import ssl
from http import HTTPStatus
from typing import TYPE_CHECKING

import aiohttp
import certifi
from aiohttp import ClientResponse, ContentTypeError
from aiolimiter import AsyncLimiter

from cyberdrop_dl.clients.download_client import DownloadClient
from cyberdrop_dl.clients.errors import DownloadFailure, DDOSGuardFailure, ScrapeFailure
from cyberdrop_dl.clients.scraper_client import ScraperClient
from cyberdrop_dl.utils.utilities import CustomHTTPStatus

if TYPE_CHECKING:
    from cyberdrop_dl.managers.manager import Manager


class ClientManager:
    """Creates a 'client' that can be referenced by scraping or download sessions"""
    def __init__(self, manager: Manager):
        self.manager = manager

        self.connection_timeout = manager.config_manager.global_settings_data['Rate_Limiting_Options']['connection_timeout']
        self.read_timeout = manager.config_manager.global_settings_data['Rate_Limiting_Options']['read_timeout']
        self.rate_limit = manager.config_manager.global_settings_data['Rate_Limiting_Options']['rate_limit']
        self.download_delay = manager.config_manager.global_settings_data['Rate_Limiting_Options']['download_delay']
        self.user_agent = manager.config_manager.global_settings_data['General']['user_agent']
        self.verify_ssl = not manager.config_manager.global_settings_data['General']['allow_insecure_connections']
        self.simultaneous_per_domain = manager.config_manager.global_settings_data['Rate_Limiting_Options']['max_simultaneous_downloads_per_domain']

        self.ssl_context = ssl.create_default_context(cafile=certifi.where()) if self.verify_ssl else False
        self.cookies = aiohttp.CookieJar(quote_cookie=False)
        self.proxy = manager.config_manager.global_settings_data['General']['proxy'] if not manager.args_manager.proxy else manager.args_manager.proxy
        self.flaresolverr = manager.config_manager.global_settings_data['General']['flaresolverr'] if not manager.args_manager.flaresolverr else manager.args_manager.flaresolverr

        self.domain_rate_limits = {
            "bunkrr": AsyncLimiter(5, 1),
            "cyberdrop": AsyncLimiter(5, 1),
            "coomer": AsyncLimiter(1, 1),
            "kemono": AsyncLimiter(1, 1),
            "pixeldrain": AsyncLimiter(10, 1),
            "other": AsyncLimiter(25, 1)
        }
        
        self.download_spacer = {'bunkr': 0.5, 'bunkrr': 0.5, 'cyberdrop': 0, 'cyberfile': 0, "pixeldrain": 0, "coomer": 0.5, "kemono": 0.5}

        self.global_rate_limiter = AsyncLimiter(self.rate_limit, 1)
        self.session_limit = asyncio.Semaphore(50)
        self.download_session_limit = asyncio.Semaphore(self.manager.config_manager.global_settings_data['Rate_Limiting_Options']['max_simultaneous_downloads'])

        self.scraper_session = ScraperClient(self)
        self.downloader_session = DownloadClient(manager, self)

    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""

    async def get_downloader_spacer(self, key: str) -> float:
        """Returns the download spacer for a domain"""
        if key in self.download_spacer:
            return self.download_spacer[key]
        return 0.1

    async def get_rate_limiter(self, domain: str) -> AsyncLimiter:
        """Get a rate limiter for a domain"""
        if domain in self.domain_rate_limits:
            return self.domain_rate_limits[domain]
        return self.domain_rate_limits["other"]

    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""

    async def check_http_status(self, response: ClientResponse, download: bool = False) -> None:
        """Checks the HTTP status code and raises an exception if it's not acceptable"""
        status = response.status
        headers = response.headers

        if download:
            if headers.get('ETag') in ['"eb669b6362e031fa2b0f1215480c4e30"', '"a9e4cee098dc6f1e09ec124299f26b30"']:
                raise DownloadFailure(status="Bunkr Maintenance", message="Bunkr under maintenance")
            if headers.get('ETag') == '"d835884373f4d6c8f24742ceabe74946"':
                raise DownloadFailure(status=HTTPStatus.NOT_FOUND, message="Imgur image has been removed")
            if headers.get('ETag') == '"65b7753c-528a"':
                raise DownloadFailure(status=HTTPStatus.NOT_FOUND, message="SC Scrape Image")

        if HTTPStatus.OK <= status < HTTPStatus.BAD_REQUEST:
            return
        
        if "gofile" in response.url.host.lower():
            try:
                JSON_Resp = await response.json()
                if "notFound" in JSON_Resp["status"]:
                    raise ScrapeFailure(404, "Does Not Exist")
            except ContentTypeError:
                pass

        if "imgur" in response.url.host.lower():
            try:
                JSON_Resp = await response.json()
                if "status" in JSON_Resp:
                    raise ScrapeFailure(JSON_Resp['status'], JSON_Resp['data']['error'])
            except ContentTypeError:
                pass

        try:
            phrase = HTTPStatus(status).phrase
        except ValueError:
            phrase = "Unknown"

        response_text = await response.text()
        if "<title>DDoS-Guard</title>" in response_text:
            raise DDOSGuardFailure(status="DDOS-Guard", message="DDoS-Guard detected")

        if not headers.get('Content-Type'):
            raise DownloadFailure(status=CustomHTTPStatus.IM_A_TEAPOT, message="No content-type in response header")

        raise DownloadFailure(status=status, message=f"HTTP status code {status}: {phrase}")



import asyncio
import copy
import json
from dataclasses import field

from cyberdrop_dl import __version__
from cyberdrop_dl.managers.args_manager import ArgsManager
from cyberdrop_dl.managers.client_manager import ClientManager
from cyberdrop_dl.managers.db_manager import DBManager
from cyberdrop_dl.managers.cache_manager import CacheManager
from cyberdrop_dl.managers.config_manager import ConfigManager
from cyberdrop_dl.managers.download_manager import DownloadManager
from cyberdrop_dl.managers.log_manager import LogManager
from cyberdrop_dl.managers.path_manager import PathManager
from cyberdrop_dl.managers.progress_manager import ProgressManager
from cyberdrop_dl.utils.args import config_definitions
from cyberdrop_dl.utils.dataclasses.supported_domains import SupportedDomains
from cyberdrop_dl.utils.transfer.first_time_setup import TransitionManager
from cyberdrop_dl.utils.utilities import log


class Manager:
    def __init__(self):
        self.args_manager: ArgsManager = ArgsManager()
        self.cache_manager: CacheManager = CacheManager(self)
        self.path_manager: PathManager = field(init=False)
        self.config_manager: ConfigManager = field(init=False)
        self.log_manager: LogManager = field(init=False)
        self.db_manager: DBManager = field(init=False)
        self.client_manager: ClientManager = field(init=False)
        self.download_manager: DownloadManager = field(init=False)
        self.progress_manager: ProgressManager = field(init=False)

        self.first_time_setup: TransitionManager = TransitionManager(self)

        self._loaded_args_config: bool = False
        self._made_portable: bool = False

        self.task_group: asyncio.TaskGroup = field(init=False)
        self.task_list: list = []
        self.scrape_mapper = field(init=False)
        
        self.vi_mode: bool = None

    def startup(self) -> None:
        """Startup process for the manager"""
        self.args_startup()

        if not self.args_manager.appdata_dir:
            self.first_time_setup.startup()

        self.path_manager = PathManager(self)
        self.path_manager.pre_startup()

        self.cache_manager.startup(self.path_manager.cache_dir / "cache.yaml")
        self.config_manager = ConfigManager(self)
        self.config_manager.startup()
        self.vi_mode = self.config_manager.global_settings_data['UI_Options']['vi_mode'] if self.args_manager.vi_mode is None else self.args_manager.vi_mode

        self.path_manager.startup()
        self.log_manager = LogManager(self)

    def args_startup(self) -> None:
        """Start the args manager"""
        if not self.args_manager.parsed_args:
            self.args_manager.startup()

    """~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"""

    async def async_startup(self) -> None:
        """Async startup process for the manager"""
        await self.args_consolidation()
        await self.args_logging()

        if not isinstance(self.db_manager, DBManager):
            self.db_manager = DBManager(self, self.path_manager.history_db)
            self.db_manager.ignore_history = self.config_manager.settings_data['Runtime_Options']['ignore_history']
            await self.db_manager.startup()
        if not isinstance(self.client_manager, ClientManager):
            self.client_manager = ClientManager(self)
        if not isinstance(self.download_manager, DownloadManager):
            self.download_manager = DownloadManager(self)
        self.progress_manager = ProgressManager(self)
        await self.progress_manager.startup()

        # set files from args
        from cyberdrop_dl.utils.utilities import MAX_NAME_LENGTHS
        MAX_NAME_LENGTHS['FILE'] = int(self.config_manager.global_settings_data['General']['max_file_name_length'])
        MAX_NAME_LENGTHS['FOLDER'] = int(self.config_manager.global_settings_data['General']['max_folder_name_length'])

    async def args_consolidation(self) -> None:
        """Consolidates runtime arguments with config values"""
        for arg in self.args_manager.parsed_args:
            if arg in config_definitions.settings['Download_Options']:
                if self.args_manager.parsed_args[arg] != config_definitions.settings['Download_Options'][arg]:
                    self.config_manager.settings_data['Download_Options'][arg] = self.args_manager.parsed_args[arg]
            elif arg in config_definitions.settings['File_Size_Limits']:
                if self.args_manager.parsed_args[arg] != config_definitions.settings['File_Size_Limits'][arg]:
                    self.config_manager.settings_data['File_Size_Limits'][arg] = self.args_manager.parsed_args[arg]
            elif arg in config_definitions.settings['Ignore_Options']:
                if self.args_manager.parsed_args[arg] != config_definitions.settings['Ignore_Options'][arg]:
                    self.config_manager.settings_data['Ignore_Options'][arg] = self.args_manager.parsed_args[arg]
            elif arg in config_definitions.settings['Runtime_Options']:
                if self.args_manager.parsed_args[arg] != config_definitions.settings['Runtime_Options'][arg]:
                    self.config_manager.settings_data['Runtime_Options'][arg] = self.args_manager.parsed_args[arg]

    async def args_logging(self) -> None:
        """Logs the runtime arguments"""
        forum_xf_cookies_provided = {}
        forum_credentials_provided = {}

        for forum in SupportedDomains.supported_forums_map.values():
            if self.config_manager.authentication_data["Forums"][f"{forum}_xf_user_cookie"]:
                forum_xf_cookies_provided[f"{forum} XF Cookie Provided"] = True
            else:
                forum_xf_cookies_provided[f"{forum} XF Cookie Provided"] = False

            if self.config_manager.authentication_data["Forums"][f"{forum}_username"] and self.config_manager.authentication_data["Forums"][f"{forum}_password"]:
                forum_credentials_provided[f"{forum} Credentials Provided"] = True
            else:
                forum_credentials_provided[f"{forum} Credentials Provided"] = False

        gofile_credentials_provided = bool(self.config_manager.authentication_data["GoFile"]["gofile_api_key"])

        imgur_credentials_provided = bool(self.config_manager.authentication_data["Imgur"]["imgur_client_id"])
        jdownloader_credentials_provided = False

        if self.config_manager.authentication_data["JDownloader"]['jdownloader_username'] and self.config_manager.authentication_data["JDownloader"]['jdownloader_password'] and self.config_manager.authentication_data["JDownloader"]['jdownloader_device']:
            jdownloader_credentials_provided = True

        pixeldrain_credentials_provided = bool(self.config_manager.authentication_data["PixelDrain"]["pixeldrain_api_key"])
        reddit_credentials_provided = False

        if self.config_manager.authentication_data["Reddit"]['reddit_personal_use_script'] and self.config_manager.authentication_data["Reddit"]['reddit_secret']:
            reddit_credentials_provided = True

        auth_provided = {
            "Forums Credentials": forum_credentials_provided,
            "Forums XF Cookies": forum_xf_cookies_provided,
            "GoFile Provided": gofile_credentials_provided,
            "Imgur Provided": imgur_credentials_provided,
            "JDownloader Provided": jdownloader_credentials_provided,
            "PixelDrain Provided": pixeldrain_credentials_provided,
            "Reddit Provided": reddit_credentials_provided
        }

        print_settings = copy.deepcopy(self.config_manager.settings_data)
        print_settings['Files']['input_file'] = str(print_settings['Files']['input_file'])
        print_settings['Files']['download_folder'] = str(print_settings['Files']['download_folder'])
        print_settings["Logs"]["log_folder"] = str(print_settings["Logs"]["log_folder"])
        print_settings['Sorting']['sort_folder'] = str(print_settings['Sorting']['sort_folder'])

        input_file = str(self.path_manager.input_file)
        download_dir = str(self.path_manager.download_dir)

        await log(f"Starting Cyberdrop-DL Process for {self.config_manager.loaded_config} Config", 10)
        await log(f"Running version {__version__}", 10)
        await log(f"Using Config: {self.config_manager.loaded_config}", 10)
        await log(f"Using Config File: {str(self.config_manager.settings)}", 10)
        await log(f"Using Input File: {input_file}", 10)
        await log(f"Using Download Folder: {download_dir}", 10)
        await log(f"Using History File: {str(self.path_manager.history_db)}", 10)

        await log(f"Using Authentication: \n{json.dumps(auth_provided, indent=4, sort_keys=True)}", 10)
        await log(f"Using Settings: \n{json.dumps(print_settings, indent=4, sort_keys=True)}", 10)
        await log(f"Using Global Settings: \n{json.dumps(self.config_manager.global_settings_data, indent=4, sort_keys=True)}", 10)

    async def close(self) -> None:
        """Closes the manager"""
        await self.db_manager.close()
